from typing import Dict, NamedTuple, Any, AsyncGenerator

from .async_live_queue import AsyncLiveQueue

default_queue_size = 5000


class SubscriptionNotification(NamedTuple):
    subscription_id: str
    data: Any
    ok: bool


class SubscriptionClosed(Exception):
    pass


class SubscriptionManager:
    """
    Tracks and handles subscription updates as JSON-RPC notifications. Maintains
    an `AsyncLiveQueue` structure to track messages for each subscription type.

    If queue structures are not properly drained old messages will be discarded.
    """

    queue_limit: int
    all_subscription_notifications: AsyncLiveQueue[SubscriptionNotification]
    subscription_notifications_by_id: Dict[
        str, AsyncLiveQueue[SubscriptionNotification]
    ]

    def __init__(self, queue_limit: int = default_queue_size):
        self.queue_limit = queue_limit
        self.all_subscription_notifications = AsyncLiveQueue(self.queue_limit)
        self.subscription_notifications_by_id = {}

    def register(self, subscription_id: str) -> None:
        """
        Registers a queue for a new subscription ID. Ignores duplicate
        subscription IDs.

        :param subscription_id: subscription ID
        """

        if subscription_id in self.subscription_notifications_by_id:
            return

        self.subscription_notifications_by_id[subscription_id] = AsyncLiveQueue(
            self.queue_limit
        )

    def delete(self, subscription_id: str) -> None:
        """
        Delete subscription ID and pushes an error to all queues listening for
        that particular message.

        :param subscription_id: subscription ID to cancel
        """
        queue = self.subscription_notifications_by_id[subscription_id]
        queue.put_nowait(SubscriptionNotification("", None, False))
        del self.subscription_notifications_by_id[subscription_id]

    async def push(self, subscription_id: str, data: Any) -> None:
        """
        Pushes a new subscription notification to the queues.

        Due to asynchronous scheduling it's possible for a message to be received
        before the subscription is registered. These messages will be dropped.
        """
        notification = SubscriptionNotification(subscription_id, data, True)
        await self.all_subscription_notifications.put(notification)
        if notification.subscription_id in self.subscription_notifications_by_id:
            await self.subscription_notifications_by_id[
                notification.subscription_id
            ].put(notification)

    async def next_all(self) -> SubscriptionNotification:
        """
        Consume notifications from any subscription notification queue.

        :return: notification
        """
        return await self.all_subscription_notifications.get()

    async def iter_all(self) -> AsyncGenerator[SubscriptionNotification, None]:
        while True:
            yield await self.all_subscription_notifications.get()

    async def next_for_id(self, subscription_id: str) -> Any:
        """
        Consume notifications from subscription queue with provided id.

        :param subscription_id: subscription queue to consume from
        :return: notification
        """
        notification = await self.subscription_notifications_by_id[
            subscription_id
        ].get()
        if not notification.ok:
            raise SubscriptionClosed()

        return notification.data

    async def iter_for_id(self, subscription_id: str) -> AsyncGenerator[Any, None]:
        queue = self.subscription_notifications_by_id[subscription_id]

        while True:
            notification = await queue.get()
            if not notification.ok:
                raise SubscriptionClosed()

            yield notification.data



import asyncio
import json
from asyncio import Future
from typing import (
    Optional,
    Any,
    NamedTuple,
    Callable,
    TypeVar,
    AsyncGenerator,
)

import websockets
import websockets.datastructures
from websockets import client as wsclient
from .response_queue import ResponseQueue
from .subscription_manager import SubscriptionManager, SubscriptionNotification
from ..types import (
    JsonRpcRequest,
    JsonRpcResponse,
    RpcMessageFormatException,
    RpcExecutionException,
)

T = TypeVar("T")


class WsRpcOpts(NamedTuple):
    encoder: Callable[[Any], str] = json.dumps
    decoder: Callable[[str], Any] = json.loads
    request_timeout_s: Optional[int] = 5
    notification_timeout_s: Optional[int] = None
    max_ws_size: int = 2**24
    headers: Optional[websockets.datastructures.HeadersLike] = None


class WsRpcConnection:
    """
    Wraps a websocket connection to serve as an RPC interface.

    Handles all reading and writing from a websockets RPC server with
    streaming as notifications support.

    Enhancement TODOs:
    - reconnection logic
    """

    uri: str
    opts: WsRpcOpts

    # internal state
    _ws: Optional[wsclient.WebSocketClientProtocol] = None
    _request_id: int = 1
    _subscription_manager: SubscriptionManager = SubscriptionManager()
    _response_queue: ResponseQueue = ResponseQueue()

    # track tasks
    _receiver_task: Optional[Future] = None
    _status_task: Optional[Future] = None

    def __init__(self, uri: str, opts: Optional[WsRpcOpts] = None):
        if opts is None:
            opts = WsRpcOpts()

        self.uri = uri
        self.opts = opts

    async def __aenter__(self) -> "WsRpcConnection":
        await self.connect()
        return self

    def __await__(self):
        yield from self.connect().__await__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def connect(self):
        """
        Connects the websocket and creates all relevant tasks managing
        websocket message and statuses.
        """
        self._ws = await self.connect_ws()
        self._receiver_task = asyncio.create_task(self._read_loop())
        self._status_task = asyncio.create_task(self._ensure_ws())

    async def connect_ws(self) -> wsclient.WebSocketClientProtocol:
        return await websockets.connect(
            self.uri,
            max_size=self.opts.max_ws_size,
            extra_headers=self.opts.headers
        )

    async def call(
        self, method_name: str, params: Any, *, request_id: Optional[str] = None
    ) -> Any:
        """
        Executes an RPC call. Dispatches a method call with request ID to RPC
        server, and waits for the matching response. `call` times out based
        on the initial options configuration.

        :param method_name: method to call
        :param params: method parameters
        :param request_id: request ID to override (should be skipped in most cases)
        :return: RPC response
        """
        if request_id is None:
            request_id = str(self._request_id)
            self._request_id += 1

        ws = self._ws
        if ws is None:
            raise WsNotConnected()

        rpc_request = JsonRpcRequest(request_id, method_name, params)
        await ws.send(rpc_request.to_jsons(encoder=self.opts.encoder))

        rpc_response = await self._wait_for_rpc_response(request_id)
        return rpc_response.unwrap()

    async def subscribe(
        self, channel: str, options: Optional[Any] = None, method_name: str = "subscribe"
    ) -> str:
        """
        Subscribes to an JSONRPC stream. Expects to receive back a subscription ID
        identifying the stream, and notifications from the server with that
        subscription ID attached.

        :param channel: name of channel to subscribe to
        :param options: stream options
        :param method_name: RPC method name for subscribing
        :return: subscription ID
        """
        if options is None:
            options = {}

        result = await self.call(method_name, [channel, options])
        if not isinstance(result, str):
            raise RpcMessageFormatException(
                "subscribe message did not return correct format"
            )

        self._subscription_manager.register(result)
        return result

    async def unsubscribe(self, subscription_id: str, method_name: str = "unsubscribe"):
        """
        Unsubscribes from a JSONRPC stream. All future messages from that stream
        are discarded.

        :param subscription_id: subscription ID to delete
        :param method_name: RPC method name for unsubscribing
        """
        response = await self.call(method_name, [subscription_id])
        if response is not True:
            raise RpcExecutionException(
                "unsubscribe call did not return True or error"
            )

        self._subscription_manager.delete(subscription_id)

    async def next_notification(self) -> SubscriptionNotification:
        """
        Returns the next notification from any subscription ID.

        :return: subscription notification
        """
        task = asyncio.create_task(self._subscription_manager.next_all())
        return await self._wait_for_message(
            task, self.opts.notification_timeout_s
        )

    async def notifications(
        self,
    ) -> AsyncGenerator[SubscriptionNotification, None]:
        """
        Returns an async generator for all notifications on the websocket connection

        :return: async generator for all notifications
        """
        async for notification in self._subscription_manager.iter_all():
            yield notification

    async def next_notification_for_id(self, subscription_id: str) -> Any:
        """
        Returns the next notification for the specified subscription ID

        :param subscription_id: filters for specific notification set
        :return: next notification
        """
        task = asyncio.create_task(
            self._subscription_manager.next_for_id(subscription_id)
        )
        return await self._wait_for_message(
            task, self.opts.notification_timeout_s
        )

    async def notifications_for_id(
        self, subscription_id: str
    ) -> AsyncGenerator[Any, None]:
        """
        Returns an async generator for all notifications for a specific subscription ID
        :param subscription_id: filters for a specific notification set
        :return: async generator for notifications
        """
        async for notification in self._subscription_manager.iter_for_id(
            subscription_id
        ):
            yield notification

    async def close(self):
        """
        Closes websocket connection gracefully.

        :return:
        """
        ws = self._ws
        assert ws is not None

        await ws.close()

    async def _read_loop(self):
        ws = self._ws
        assert ws is not None

        async for message in ws:
            # response type
            try:
                response_message = JsonRpcResponse().from_jsons(
                    message, decoder=self.opts.decoder
                )
            except RpcMessageFormatException as e:
                pass
            else:
                await self._response_queue.put(response_message)
                continue

            # notification type
            try:
                subscription_message = JsonRpcRequest().from_jsons(
                    message, decoder=self.opts.decoder
                )
                params = subscription_message.params
                assert isinstance(params, dict)
                await self._subscription_manager.push(
                    params["subscription"], params["result"]
                )
            except RpcMessageFormatException as e:
                await ws.close(3000, "message did not conform to JSON-RPC spec")

    async def _ensure_ws(self):
        ws = self._ws
        await ws.wait_closed()

    async def _wait_for_rpc_response(self, request_id: str) -> JsonRpcResponse:
        task = asyncio.create_task(self._response_queue.get(request_id))
        return await self._wait_for_message(task, self.opts.request_timeout_s)

    async def _wait_for_message(
        self, message_task: "Future[T]", timeout: Optional[int] = None
    ) -> T:
        status_task = self._status_task
        if status_task is None:
            raise WsNotConnected()

        await asyncio.wait(
            [status_task, message_task],
            return_when=asyncio.FIRST_COMPLETED,
            timeout=timeout,
        )

        if message_task.done():
            return message_task.result()

        if status_task.done():
            raise WsNotConnected()

        message_task.cancel()
        raise WsRpcTimedOut()


class WsException(Exception):
    pass


class WsNotConnected(WsException):
    pass


class WsRpcTimedOut(WsException):
    pass



# exceptions raised by RPC client implementation in response to inconsistent behavior from server


class RpcMessageFormatException(Exception):
    pass


class RpcExecutionException(Exception):
    pass



import json
from typing import Optional, Any, Dict, Union, TypeVar, Generic

import humps

from .client_error import RpcMessageFormatException
from .case import Case
from .constants import *
from .server_error import RpcError
from .. import utils

T = TypeVar("T")


class JsonRpcResponse(Generic[T]):
    id: Optional[str]
    result: Optional[T]
    error: Optional[RpcError]
    jsonrpc_version: str = JSONRPC_VERSION

    def __init__(
        self,
        request_id: Optional[str] = None,
        result: Optional[T] = None,
        error: Optional[RpcError] = None,
    ) -> None:
        self.id = request_id
        self.result = result
        self.error = error

    def __str__(self) -> str:
        return f"JsonRpcResponse<{self.to_jsons()}>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, JsonRpcResponse):
            return (
                self.id == o.id
                and self.result == o.result
                and self.error == o.error
                and self.jsonrpc_version == o.jsonrpc_version
            )
        else:
            return False

    def unwrap(self) -> T:
        if self.error is None:
            return self.result

        raise self.error

    def to_json(self, case: Case = Case.SNAKE) -> Dict[str, Any]:
        fields = {
            "jsonrpc": self.jsonrpc_version,
            "id": self.id,
        }
        result = self.result
        if result is not None:
            if case == Case.CAMEL and isinstance(result, dict):
                fields["result"] = humps.camelize(result)
            else:
                fields["result"] = result

        error = self.error
        if error is not None:
            fields["error"] = error.to_json()
        return fields

    def to_jsons(self, case: Case = Case.SNAKE, encoder=json.dumps) -> str:
        return encoder(self.to_json(case))

    def from_json(self, payload: Dict[str, Any]) -> "JsonRpcResponse":
        if "id" not in payload:
            raise RpcMessageFormatException(
                "response did not contain a request id"
            )

        if not ("result" not in payload) ^ ("error" not in payload):
            raise RpcMessageFormatException(
                "response contained neither (or both) result and error"
            )

        self.id = payload.get("id", None)
        self.result = payload.get("result", None)
        self.error = utils.optional_map(
            payload.get("error", None), RpcError.from_json
        )
        return self

    def from_jsons(
        self, payload: Union[bytes, str], decoder=json.loads
    ) -> "JsonRpcResponse":
        return self.from_json(decoder(payload))



from typing import Optional, TypeVar, Callable

T = TypeVar("T")
R = TypeVar("R")


def optional_map(val: Optional[T], mapper: Callable[[T], R]) -> Optional[R]:
    if val is None:
        return val
    else:
        return mapper(val)


def or_else(val: Optional[T], default: T) -> T:
    if val is None:
        return default
    else:
        return val




import asyncio
from typing import TypeVar, Generic

T = TypeVar("T")


class AsyncLiveQueue(asyncio.Queue, Generic[T]):
    """
    asyncio.Queue with liveliness characteristics.

    When the queue is full, old items are ejected in favor of newer entries.
    """

    def __init__(self, max_size: int):
        super().__init__(max_size)

    async def put(self, item: T) -> None:
        if self.full():
            self.get_nowait()
        return self.put_nowait(item)



import json
from typing import Optional, Any, Dict, Union

import humps

from .client_error import RpcMessageFormatException
from .case import Case
from .constants import *


class JsonRpcRequest:
    id: Optional[str]
    method: str
    params: Any
    jsonrpc_version: str = JSONRPC_VERSION

    def __init__(
        self,
        request_id: Optional[str] = None,
        method: str = "",
        params: Any = None,
    ) -> None:
        self.id = request_id
        self.method = method
        self.params = params

    def __str__(self) -> str:
        return f"JsonRpcRequest<{self.to_json()}>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, JsonRpcRequest):
            return (
                self.id == o.id
                and self.method == o.method
                and self.params == o.params
                and self.jsonrpc_version == o.jsonrpc_version
            )
        else:
            return False

    def to_json(self, case: Case = Case.SNAKE) -> Dict[str, Any]:
        params = self.params
        if params is not None and case == Case.CAMEL:
            params = humps.camelize(params)

        return {
            "jsonrpc": self.jsonrpc_version,
            "id": self.id,
            "method": self.method,
            "params": params,
        }

    # TODO: needs type annotation for encoder (probably module?)
    def to_jsons(self, case: Case = Case.SNAKE, encoder=json.dumps) -> str:
        json_dict = self.to_json(case)
        return encoder(json_dict)

    def from_json(self, payload: Dict[str, Any]) -> "JsonRpcRequest":
        self.id = payload.get("id", None)
        if "method" not in payload:
            raise RpcMessageFormatException("request does not contain a method")

        self.method = payload["method"]
        self.params = payload.get("params", None)
        return self

    def from_jsons(self, payload: Union[bytes, str], decoder=json.loads) -> "JsonRpcRequest":
        return self.from_json(decoder(payload))



from .request import JsonRpcRequest
from .response import JsonRpcResponse
from .client_error import RpcMessageFormatException, RpcExecutionException
from .server_error import (
    RpcError,
    RpcParseError,
    RpcInternalError,
    RpcInvalidRequest,
    RpcInvalidParams,
    RpcMethodNotFound,
)
from .case import Case

__all__ = [
    "Case",
    "JsonRpcRequest",
    "JsonRpcResponse",
    "RpcError",
    "RpcParseError",
    "RpcInternalError",
    "RpcInvalidParams",
    "RpcInvalidRequest",
    "RpcMethodNotFound",
    "RpcMessageFormatException",
    "RpcExecutionException"
]



from .types import (
    Case,
    JsonRpcRequest,
    JsonRpcResponse,
    RpcError,
    RpcParseError,
    RpcInternalError,
    RpcInvalidRequest,
    RpcInvalidParams,
    RpcMethodNotFound,
)
from .ws import (
    WsRpcConnection,
    WsRpcOpts,
    WsException,
    WsNotConnected,
    WsRpcTimedOut,
)


__all__ = [
    "Case",
    "JsonRpcRequest",
    "JsonRpcResponse",
    "RpcError",
    "RpcParseError",
    "RpcInternalError",
    "RpcInvalidParams",
    "RpcInvalidRequest",
    "RpcMethodNotFound",
    "WsRpcConnection",
    "WsRpcOpts",
    "WsException",
    "WsNotConnected",
    "WsRpcTimedOut",
]
